/***********************************************************************\
* vov is a tool that emulates the behavior of a Von-Neumann machine.    *
* Copyright (C) 2001-2008 Davide Scola <davide.scola@gmail.com>         *
*                                                                       *
* This file is part of vov.                                             *
*                                                                       *
* vov is free software; you can  redistribute it and/or modify it under *
* the terms of the GNU General  Public License as published by the Free *
* Software Foundation;  either version  2 of the  License, or  (at your *
* option) any later version.                                            *
*                                                                       *
* vov is  distributed in the hope  that it will be  useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or *
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License *
* for more details.                                                     *
*                                                                       *
* You should  have received a  copy of  the GNU General  Public License *
* along  with vov;  if  not,  write to  the  Free Software  Foundation, *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.    *
\***********************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include "allocator.h"

#if ! HAVE_MALLOC
# if HAVE_MALLOC_H
#  include <malloc.h>
# else
#  include "rpl_malloc.h"
# endif
#endif

#ifdef DEBUG
# include "line.h"
# include "buffer.h"
# include "utils.h"
# include "human_out.h"
#endif


/*
 * global variable 
 */
static allocator_t memory;


static inline ALWAYS_INLINE page_t *
p_alloc(unsigned long int size)
{
    page_t         *y;

    if (NULL != (y = malloc(size))) {
	y->x = (y + 1);
	y->p = 0;
	y->a = size - sizeof(page_t);
	y->next = NULL;
    }

    return y;
}


/*
 * add a new page to the list; return the new
 * page or NULL on error
 */
static inline ALWAYS_INLINE page_t *
add_page(unsigned long int size)
{
    page_t         *y;
    register int    wanted;

    /*
     * catch integer overflow 
     */
    if (size + sizeof(page_t) < sizeof(page_t))
	return NULL;

    size += sizeof(page_t);

    /*
     * this always ensure that wanted is greather than or equal to size 
     */
    wanted = ((size > MEMORY_PAGE_SIZE) ? size : MEMORY_PAGE_SIZE);

    if (NULL != (y = p_alloc(wanted))) {
	if (NULL == memory.head)
	    memory.tail = memory.head = y;

	else
	    memory.tail = memory.tail->next = y;
    }

    return y;
}


/*
 * looking for the first page that can held the requested size;
 * if no page is found, a new one is created
 */
static inline ALWAYS_INLINE page_t *
find_page(unsigned long int size)
{
    /*
     * since vov is using this allocator only for the list nodes
     * (therefore on data with same size) if we have enough free bytes
     * they always are in the last page (so we don't need to scan the
     * entire list) 
     */
    if ((NULL != memory.tail)
	&& ((memory.tail->a - memory.tail->p) >= size))
	return memory.tail;

    return add_page(size);
}


static inline ALWAYS_INLINE void *
grows_memory_size(unsigned long int size)
{
    page_t         *y;

    if (NULL == (y = find_page(size)))
	return NULL;

    /*
     * update current pointer 
     */
    y->p += size;

    return ((char *) (y->x) + y->p - size);
}


void           *
vov_malloc(unsigned long size)
{
    return grows_memory_size(size);
}


/*
 * free all pages
 */
static void
free_pages(page_t * page)
{
    if (NULL != page) {
	/*
	 * free next page 
	 */
	free_pages(page->next);

	/*
	 * free list 
	 */
	free(page);
    }
}


/*
 * API to access the allocator 
 */
void
vov_free(void)
{
    free_pages(memory.head);
}


#ifdef DEBUG
/*
 * print memory status 
 */
void
print_memory(void)
{
    char            stra[FMT_ULONG];
    char            strb[FMT_ULONG];
    char            strc[FMT_ULONG];
    page_t         *y;
    int             i;

    msg("+++++++ This is the allocator dump +++++++\n");

    for (i = 0, y = memory.head; y != NULL; ++i, y = y->next) {
	stra[stringify(stra, i)] = '\0';
	strb[stringify(strb, y->a)] = '\0';
	strc[stringify(strc, y->p)] = '\0';

	msg("- page ", stra, " of size ", strb,
	    " bytes with current pointer at ", strc, " byte(s).\n");
    }

    msg("++++++++++++++++++++++++++++++++++++++++++\n");
}
#endif				/* DEBUG */
