/***********************************************************************\
* vov is a tool that emulates the behavior of a Von-Neumann machine.    *
* Copyright (C) 2001-2008 Davide Scola <davide.scola@gmail.com>         *
*                                                                       *
* This file is part of vov.                                             *
*                                                                       *
* vov is free software; you can  redistribute it and/or modify it under *
* the terms of the GNU General  Public License as published by the Free *
* Software Foundation;  either version  2 of the  License, or  (at your *
* option) any later version.                                            *
*                                                                       *
* vov is  distributed in the hope  that it will be  useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or *
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License *
* for more details.                                                     *
*                                                                       *
* You should  have received a  copy of  the GNU General  Public License *
* along  with vov;  if  not,  write to  the  Free Software  Foundation, *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.    *
\***********************************************************************/

/*****************************************************************\
*                           CREDITS                               *
*                                                                 *
* This code  is based on  libowfat (http://www.fefe.de/libowfat/) *
* written by Felix von Leitner and released under the GNU General *
* Public License published by the Free Software Foundation.       *
* In short, libowfat reimplements libdjb under GPL.               *
*                                                                 *
* Since I needed to change some little things and I don't want to *
* add any extra  no standard dependence to vov, I wrote this file *
\*****************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef __MINGW32__
# include <stdio.h>             /* STD(IN|OUT|ERR)_FILENO and NULL here */
#endif

#include <errno.h>
#include <unistd.h>
#include <stdarg.h>
#include "line.h"
#include "utils.h"
#include "buffer.h"


/*
 * a flag that will be used in signal handlers 
 */
int             exitasap;


/*
 * global variables; initialize standard buffers 
 */
char            stdin_buffer_space[1024];
static buffer_t stdin_buffer =
    { stdin_buffer_space, 0, 0, sizeof(stdin_buffer_space), STDIN_FILENO,
(cb_t) read };
buffer_t       *buffer_0 = &stdin_buffer;

char            stdout_buffer_space[1024];
static buffer_t stdout_buffer =
    { stdout_buffer_space, 0, 0, sizeof(stdout_buffer_space),
STDOUT_FILENO, (cb_t) write };
buffer_t       *buffer_1 = &stdout_buffer;

char            stderr_buffer_space[1024];
static buffer_t stderr_buffer =
    { stderr_buffer_space, 0, 0, sizeof(stderr_buffer_space),
STDERR_FILENO, (cb_t) write };
buffer_t       *buffer_2 = &stderr_buffer;


/******************************** INPUT ********************************/

/*
 * make one read on the given file descriptor and fill the buffer
 */
static inline ALWAYS_INLINE int
buffer_read(cb_t op, int fd, char *buf, unsigned long int len)
{
    int             w;

    while (1 != exitasap) {
	if ((w = op(fd, buf, len)) < 0) {
	    if (errno == EINTR)
		continue;
	}

	break;
    }

    /*
     * if exitasap is 1, simulate an EOF 
     */
    return ((1 == exitasap) ? 0 : w);
}


/*
 * check if we used the entire buffer; if so then make
 * another read; in any case returns the current position
 * (or -1 on error or 0 on EOF)
 */
static inline ALWAYS_INLINE int
buffer_feed(buffer_t * buff)
{
    int             w;

    if (buff->p == buff->n) {
	if ((w = buffer_read(buff->op, buff->fd, buff->x, buff->a)) < 0)
	    return -1;

	buff->n = w;
	buff->p = 0;
    }

    return (buff->n - buff->p);
}


/*
 * extract the current character on the buffer making an
 * additional read if needed
 */
static inline ALWAYS_INLINE int
buffer_getc(buffer_t * buff, char *x)
{
    register int    blen;

    if (buff->p == buff->n) {
	if ((blen = buffer_feed(buff)) <= 0)
	    return blen;
    }

    *x = buff->x[buff->p];
    ++buff->p;

    return 1;
}


/*
 * in short, this function fills the line 'till the requested token
 * was found; returns -1 on error, 0 on EOF, 1 otherwise
 */
static inline ALWAYS_INLINE int
buffer_get_token(buffer_t * buff, line_t * ln, const char *token,
		 unsigned long int toklen)
{
    char            x;

    ln->len = 0;		/* init current line len */

    while (1) {
	if (!safe_grows_line(ln, 1)) {
	    errno = ENOMEM;
	    return -1;
	}

	switch (buffer_getc(buff, &x)) {
	case -1:
	    return -1;
	case 0:
	    return 0;
	}

	line_tail(ln, &x);

	if (byte_chr(token, toklen, x) < toklen)
	    break;
    }

    return 1;
}


/*
 * get an entire line from the buffer; an entire line means
 * a sequence of bytes terminated by '\n' (line feed)
 */
int
get_stdin_line(line_t * ln)
{
    return buffer_get_token(buffer_0, ln, "\n", 1);
}


/******************************** OUTPUT ********************************/

/*
 * write len bytes of buf on the given descriptor
 */
static inline ALWAYS_INLINE int
buffer_write(cb_t op, int fd, const char *buf, unsigned long int len)
{
    int             w;

    while ((1 != exitasap) && (0 != len)) {
	if ((w = op(fd, buf, len)) < 0) {
	    if (errno == EINTR)
		continue;

	    return -1;
	}

	buf += w;
	len -= w;
    }

    return 0;
}


/*
 * flush the given buffer; like the standard fflush, it is defined only for
 * output buffers; return -1 on error, 0 otherwise.
 */
static inline ALWAYS_INLINE int
buffer_flush(buffer_t * buff)
{
    register int    p = buff->p;

    if (!p)
	return 0;		/* buffer already empty */

    buff->p = 0;

    return buffer_write(buff->op, buff->fd, buff->x, p);
}


/*
 * put the given string on the buffer; if it doesn't fit
 * then flush the buffer first
 */
static inline ALWAYS_INLINE int
buffer_put(buffer_t * b, const char *buf, unsigned long int len)
{
    if (len > b->a - b->p) {	/* doesn't fit */
	if (buffer_flush(b) == -1)
	    return -1;

	if (len > b->a) {
	    if (buffer_write(b->op, b->fd, buf, len) < 0)
		return -1;

	    return 0;
	}
    }

    byte_copy(b->x + b->p, len, buf);
    b->p += len;
    return 0;
}


/*
 * print the given string on the given buffer; returns
 * -1 on error, 0 otherwise.
 */
static inline ALWAYS_INLINE int
buffer_puts(buffer_t * buff, const char *s)
{
    return buffer_put(buff, s, str_len(s));
}


/*
 * print the given strings list on the given buffer
 * WARNING: list _must_ be NULL terminated (orginal function
 *          was called `buffer_putm_internal_flush')
 */
int
buffer_putm_flush(buffer_t * b, ...)
{
    int             r = 0;
    va_list         a;
    const char     *s;

    va_start(a, b);

    while (NULL != (s = va_arg(a, const char *))) {
	if (-1 == buffer_puts(b, s)) {
	    r = -1;
	    break;
	}
    }

    va_end(a);
    buffer_flush(b);
    return r;
}
