/***********************************************************************\
* vov is a tool that emulates the behavior of a Von-Neumann machine.    *
* Copyright (C) 2001-2008 Davide Scola <davide.scola@gmail.com>         *
*                                                                       *
* This file is part of vov.                                             *
*                                                                       *
* vov is free software; you can  redistribute it and/or modify it under *
* the terms of the GNU General  Public License as published by the Free *
* Software Foundation;  either version  2 of the  License, or  (at your *
* option) any later version.                                            *
*                                                                       *
* vov is  distributed in the hope  that it will be  useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or *
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License *
* for more details.                                                     *
*                                                                       *
* You should  have received a  copy of  the GNU General  Public License *
* along  with vov;  if  not,  write to  the  Free Software  Foundation, *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.    *
\***********************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef __MINGW32__
# include <stdlib.h>            /* _exit here */
#endif

#include <unistd.h>
#include <signal.h>
#include "parachute.h"


/*
 * global variables 
 */
extern int      exitasap;
extern int      signum;


/*
 * set func as signal handler of sig 
 */
static inline ALWAYS_INLINE void
catch_signal(int sig, handler_t func)
{
#ifdef HAVE_SIGACTION
    struct sigaction handle;

    handle.sa_handler = func;	/* set the signal handler */
    handle.sa_flags = 0;	/* do not restart syscall */
    (void) sigemptyset(&handle.sa_mask);	/* initializes sa_mask to
						 * empty, excluding all
						 * signals */
    (void) sigaction(sig, &handle, (struct sigaction *) 0);	/* add
								 * handler 
								 */
#else
    (void) signal(sig, func);	/* try to use this ANSI API */
#endif
}


/*
 * catch some signals raised on bug conditions and terminate the
 * process immediately with an exit code of 81; if you see this
 * return code, please drop me a mail
 */
static RETSIGTYPE
bug_catch(int x)
{
    x = x;			/* make the compiler happy */
    _exit(81);
}


static inline ALWAYS_INLINE void
catch_bugs(void)
{
#ifdef SIGILL			/* Illegal Instruction (Core) */
    catch_signal(SIGILL, bug_catch);
#endif

#ifdef SIGABRT			/* Abort signal from abort (Core) */
    catch_signal(SIGABRT, bug_catch);
#endif

#ifdef SIGFPE			/* Floating point exception (Core) */
    catch_signal(SIGFPE, bug_catch);
#endif

#ifdef SIGBUS			/* Bus error - bad memory access (Core) */
    catch_signal(SIGBUS, bug_catch);
#endif

#ifdef SIGSEGV			/* Invalid memory reference (Core) */
    catch_signal(SIGSEGV, bug_catch);
#endif

#ifdef SIGSYS			/* Bad argument to routine (Core) */
    catch_signal(SIGSYS, bug_catch);
#endif

#ifdef SIGEMT			/* it's not specified in POSIX: it
				 * terminate the process (Core) */
    catch_signal(SIGEMT, bug_catch);
#endif
}


/*
 * ignore some unused signals 
 */
static inline ALWAYS_INLINE void
ignore_misc(void)
{
#ifdef SIGALRM			/* Timer signal from alarm (Term) */
    catch_signal(SIGALRM, SIG_IGN);
#endif

#ifdef SIGVTALRM		/* Virtual alarm clock (Term) */
    catch_signal(SIGVTALRM, SIG_IGN);
#endif

#ifdef SIGPROF			/* Profiling timer expired (Term) */
    catch_signal(SIGPROF, SIG_IGN);
#endif

#ifdef SIGQUIT			/* Quit from keyboard (Core) */
    catch_signal(SIGQUIT, SIG_IGN);
#endif

#ifdef SIGHUP			/* Hangup detected on controlling terminal 
				 * (Term) */
    catch_signal(SIGHUP, SIG_IGN);
#endif

#ifdef SIGXFSZ			/* File size limit exceeded (Core) */
    catch_signal(SIGXFSZ, SIG_IGN);
#endif
}


/*
 * set given handler for used signals 
 */
static inline ALWAYS_INLINE void
set_used_signals(handler_t func)
{
#ifdef SIGINT			/* Interrupt from keyboard (Term) */
    catch_signal(SIGINT, func);
#endif

#ifdef SIGTERM			/* Termination signal (Term) */
    catch_signal(SIGTERM, func);
#endif

#ifdef SIGBREAK			/* Control-break on win32 [I'll treat this 
				 * signal like SIGKILL on *nix (Term)] */
    catch_signal(SIGBREAK, func);
#endif
}


/*
 * Break only  the cpu's loop so that we do as little as possible in the
 * signal handler; it's probably not safe to do anything with standard I/O
 * streams in a signal handler - the ANSI C standard doesn't say it is.
 * This  routine should be  safe to use inside a signal handler on UNIX
 * or a console control handler on Windows, as it merely sets a flag that
 * is checked within the loop.
 */
static RETSIGTYPE
deploy_parachute(int x)
{
    signum = x;
    exitasap = 1;
}


/*
 * init signals 
 */
void
init_signals(void)
{
    /*
     * don't ignore this signal during the bootstrap or we may risk to
     * waste cpu time (quite impossible) 
     */
#ifdef SIGXCPU			/* CPU time limit exceeded (Core) */
    catch_signal(SIGXCPU, deploy_parachute);
#endif

    catch_bugs();
    ignore_misc();

    /*
     * when start and during the bootstrap process ignore these signals 
     */
    set_used_signals(SIG_IGN);
}


/*
 * handle used signal using our handler 
 */
void
install_parachute(void)
{
    set_used_signals(deploy_parachute);
}
