/***********************************************************************\
* vov is a tool that emulates the behavior of a Von-Neumann machine.    *
* Copyright (C) 2001-2008 Davide Scola <davide.scola@gmail.com>         *
*                                                                       *
* This file is part of vov.                                             *
*                                                                       *
* vov is free software; you can  redistribute it and/or modify it under *
* the terms of the GNU General  Public License as published by the Free *
* Software Foundation;  either version  2 of the  License, or  (at your *
* option) any later version.                                            *
*                                                                       *
* vov is  distributed in the hope  that it will be  useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or *
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License *
* for more details.                                                     *
*                                                                       *
* You should  have received a  copy of  the GNU General  Public License *
* along  with vov;  if  not,  write to  the  Free Software  Foundation, *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.    *
\***********************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef __MINGW32__
# include <stdio.h>             /* STD(IN|OUT|ERR)_FILENO and NULL here */
#endif

#include <stdlib.h>
#include <unistd.h>

#ifdef HAVE_STDINT_H
# include <stdint.h>
#endif

#include "line.h"
#include "buffer.h"
#include "ram.h"
#include "allocator.h"
#include "human_out.h"

#ifdef DEBUG
# include "utils.h"
#endif


/*
 * extern variables 
 */
extern unsigned long nline;


/*
 * allocate new word 
 */
static inline ALWAYS_INLINE int
create_word(ram_t ** ptr)
{
    if ((*ptr = vov_malloc(sizeof(struct word))) == NULL) {
	err("out of memory\n");
	return 0;
    }

    return 1;
}


/*
 * single left rotation; return new root 
 */
static inline ALWAYS_INLINE ram_t *
single_lrotate(ram_t * ram)
{
    ram_t          *tmp = NULL;

    tmp = ram->left;
    ram->left = tmp->right;
    tmp->right = ram;

    ram->height =
	MAX_HEIGHT(GET_HEIGHT(ram->left), GET_HEIGHT(ram->right)) + 1;
    tmp->height = MAX_HEIGHT(GET_HEIGHT(tmp->left), ram->height) + 1;
    return tmp;
}


/*
 * single right rotation; return new root 
 */
static inline ALWAYS_INLINE ram_t *
single_rrotate(ram_t * ram)
{
    ram_t          *tmp = NULL;

    tmp = ram->right;
    ram->right = tmp->left;
    tmp->left = ram;

    ram->height =
	MAX_HEIGHT(GET_HEIGHT(ram->left), GET_HEIGHT(ram->right)) + 1;
    tmp->height = MAX_HEIGHT(GET_HEIGHT(tmp->right), ram->height) + 1;
    return tmp;
}


/*
 * double left rotation; return new root 
 */
static inline ALWAYS_INLINE ram_t *
double_lrotate(ram_t * ram)
{
    ram->left = single_rrotate(ram->left);
    return single_lrotate(ram);
}


/*
 * double right rotation; return new root 
 */
static inline ALWAYS_INLINE ram_t *
double_rrotate(ram_t * ram)
{
    ram->right = single_lrotate(ram->right);
    return single_rrotate(ram);
}


/*
 * add a new word to the RAM 
 */
int
add_word(uint16_t offset, int32_t opcode, ram_t ** ram)
{
    /*
     * add new word 
     */
    if (NULL == *ram) {
	if (!create_word(ram))
	    return 0;

	(*ram)->offset = offset;
	(*ram)->opcode = opcode;
	(*ram)->height = 0;
	(*ram)->left = (*ram)->right = NULL;
    }

    /*
     * left sub-tree 
     */
    else if (offset < (*ram)->offset) {
	if (!add_word(offset, opcode, &(*ram)->left))
	    return 0;

	if (2 == (GET_HEIGHT((*ram)->left) - GET_HEIGHT((*ram)->right))) {
	    if (offset < (*ram)->left->offset)
		*ram = single_lrotate(*ram);

	    else
		*ram = double_lrotate(*ram);
	}
    }

    /*
     * right sub-tree 
     */
    else if (offset > (*ram)->offset) {
	if (!add_word(offset, opcode, &(*ram)->right))
	    return 0;

	if (2 == (GET_HEIGHT((*ram)->right) - GET_HEIGHT((*ram)->left))) {
	    if (offset > (*ram)->right->offset)
		*ram = single_rrotate(*ram);

	    else
		*ram = double_rrotate(*ram);
	}
    }

    /*
     * uh oh, a value was already set on word; update it with the new one 
     */
    else
	(*ram)->opcode = opcode;

    /*
     * update height 
     */
    (*ram)->height =
	MAX_HEIGHT(GET_HEIGHT((*ram)->left),
		   GET_HEIGHT((*ram)->right)) + 1;
    return 1;
}


ram_t          *
word_seek(uint16_t offset, ram_t * ram)
{
    if (NULL == ram)
	return NULL;

    if (offset < ram->offset)
	return word_seek(offset, ram->left);

    else if (offset > ram->offset)
	return word_seek(offset, ram->right);

    return ram;
}


/*
 * free all words 
 */
void
destroy_ram(ram_t * ram)
{
    vov_free();
    ram = NULL;			/* make the compiler happy */
}


/*
 * some debug routines 
 */
#ifdef DEBUG
static void
recursive_tree_print(ram_t * ram, int depth)
{
    /*
     * ISO C90 forbids variable-size array but who cares on a debug code?
     * :P 
     */
    char            buf[depth * 3 + 1];
    char            stra[FMT_ULONG];

    if (NULL == ram)
	return;

    recursive_tree_print(ram->right, depth + 1);

    DEBUG_MKSPACES(buf);
    stra[stringify(stra, ram->offset)] = '\0';

    msg(buf, stra, "\n");
    recursive_tree_print(ram->left, depth + 1);
}


/*
 * print (tree version) RAM 
 */
void
print_tree_ram(ram_t * ram)
{
    msg("+++++++ Tree version of your RAM +++++++\n");
    recursive_tree_print(ram, 0);
    msg("++++++++++++++++++++++++++++++++++++++++\n");
}


/*
 * print (flat version) RAM 
 */
static void
recursive_flat_print(ram_t * ram)
{
    char            stra[FMT_ULONG];
    char            strb[FMT_ULONG];

    if (NULL != ram) {
	recursive_flat_print(ram->left);

	stra[stringify(stra, ram->offset)] = '\0';
	strb[stringify(strb, ram->opcode)] = '\0';
	msg("RAM[", stra, "] = ", strb, ";\n");

	recursive_flat_print(ram->right);
    }
}


/*
 * print (flat version) RAM 
 */
void
print_flat_ram(ram_t * ram)
{
    msg("+++++++ Flat version of your RAM +++++++\n");
    recursive_flat_print(ram);
    msg("++++++++++++++++++++++++++++++++++++++++\n");
}
#endif				/* ! DEBUG */
