%{
/***********************************************************************\
* vov is a tool that emulates the behavior of a Von-Neumann machine.    *
* Copyright (C) 2001-2008 Davide Scola <davide.scola@gmail.com>         *
*                                                                       *
* This file is part of vov.                                             *
*                                                                       *
* vov is free software; you can  redistribute it and/or modify it under *
* the terms of the GNU General  Public License as published by the Free *
* Software Foundation;  either version  2 of the  License, or  (at your *
* option) any later version.                                            *
*                                                                       *
* vov is  distributed in the hope  that it will be  useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or *
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License *
* for more details.                                                     *
*                                                                       *
* You should  have received a  copy of  the GNU General  Public License *
* along  with vov;  if  not,  write to  the  Free Software  Foundation, *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.    *
\***********************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <unistd.h>             /* this header should be automatically
                                 * included; btw, just to be sure ... */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#if ! HAVE_MALLOC
# if HAVE_MALLOC_H
#  include <malloc.h>
# else
#  include "rpl_malloc.h"
# endif
#endif

#if ! HAVE_REALLOC
# include "rpl_realloc.h"
#endif

#include "line.h"
#include "buffer.h"
#include "scanner.h"
#include "parser.h"
#include "human_out.h"
#include "utils.h"

unsigned long nline = 1;
unsigned long nchars;
unsigned long nstatement;

/*
 * add these prototypes to make the compiler happy
 */
extern int yyparse(void);
static void my_fatal_error(const char msg[]);

%}

extern unsigned long int yylval;

%%
[0-9]+			{ /* catch all numbers */
			  ulongify(yytext, &yylval);
			  nchars += yyleng;
			  return NUMBER;
			}

[Rr][Aa][Mm]		{ /* memory token */
			  nchars += yyleng;
			  return MEMORY;
			}

"#"+[^\n]*\n		{ /* one line comment */
			  ++nline;
			  nchars = 0;
			  return COMMENT;
			}

[ \t\r\n]+		{ /* ignore blanks */
			  register int i;

			  for (i = 0; i < yyleng; i++) {
			    if ('\n' == *(yytext + i)) { ++nline; nchars = 0; }
			    else ++nchars;
			  }
			}

.			{ ++nchars;
			  return yytext[0];
			}
%%

/*
 * show error messages
 */
void
yyerror(const char msg[])
{
    char            strnline[FMT_ULONG];
    char            strnchars[FMT_ULONG];

    strnline[stringify(strnline, nline)] = '\0';
    strnchars[stringify(strnchars, nchars)] = '\0';

    err("line ", strnline, ":", strnchars, ": token `", yytext, "': ", msg,
        "\n");
}

/*
 * re-define fatal error function, so we can free all allocated memory
 */
static void
my_fatal_error(const char msg[])
{
    err(msg);

    /*
     * before exit, cleanup all memory
     */
    if ((0 < yyin) && (yyin != STDIN_FILENO))
        (void) close(yyin);

    yy_delete_buffer(yy_current_buffer);

    _exit(YY_EXIT_FAILURE);
}

/*
 * load program
 */
int
load_external_program(const char *fname)
{
    int             ret;

    if (NULL != fname) {
        if (-1 == (yyin = open(fname, O_RDONLY))) {
            err(fname, ": impossible to open the given script\n");
            return 0;
        }
    }

    /*
     * let's parse
     */
    ret = yyparse();

    /*
     * before return, cleanup all memory
     */
    if ((0 < yyin) && (yyin != STDIN_FILENO))
        (void) close(yyin);

    yy_delete_buffer(yy_current_buffer);

    /*
     * return the code using my convention; if ret is 0, return 1 and
     * vice-versa
     */
    return ((0 == ret) ? 1 : 0);
}
