/***********************************************************************\
* vov is a tool that emulates the behavior of a Von-Neumann machine.    *
* Copyright (C) 2001-2008 Davide Scola <davide.scola@gmail.com>         *
*                                                                       *
* This file is part of vov.                                             *
*                                                                       *
* vov is free software; you can  redistribute it and/or modify it under *
* the terms of the GNU General  Public License as published by the Free *
* Software Foundation;  either version  2 of the  License, or  (at your *
* option) any later version.                                            *
*                                                                       *
* vov is  distributed in the hope  that it will be  useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or *
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License *
* for more details.                                                     *
*                                                                       *
* You should  have received a  copy of  the GNU General  Public License *
* along  with vov;  if  not,  write to  the  Free Software  Foundation, *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.    *
\***********************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef __MINGW32__
# include <stdio.h>             /* STD(IN|OUT|ERR)_FILENO and NULL here */
#endif

#include <unistd.h>

#ifdef HAVE_STDINT_H
# include <stdint.h>
#endif

#include "line.h"
#include "buffer.h"
#include "ram.h"
#include "cpu.h"
#include "vov.h"
#include "human_out.h"
#include "utils.h"
#include "sym_table.h"
#include "options.h"


/*
 * extern variables 
 */
extern vov_t    vov;
extern uint8_t  options;
extern int      exitasap;
extern buffer_t buff;
extern line_t   line;


static inline ALWAYS_INLINE int
catch_add_bug(int32_t x, int32_t y)
{
    if (x > 0 && y > 0 && (
#if GREATER_SIGNED
			      ((uint32_t) ((uint32_t) x + (uint32_t) y) <
			       (uint32_t) x)
#else
			      (((uint32_t) x + (uint32_t) y) >
			       (uint32_t) INT32_MAX)
#endif
	))
	return OFLOW;

    else if (x < 0 && y < 0 && (
#if GREATER_NEGATIVE
				   (x == INT32_MIN || y == INT32_MIN) ||
#if GREATER_SIGNED
				   (((uint32_t) (-x) + (uint32_t) (-y) !=
				     0)
				    && (uint32_t) ((uint32_t) (-x) +
						   (uint32_t) (-y))
				    < (uint32_t) (-x))
#else
				   (((uint32_t) (-x) + (uint32_t) (-y))
				    > ((uint32_t) 1 +
				       (uint32_t) INT32_MAX))
#endif
#else
#if GREATER_SIGNED
				   ((uint32_t)
				    ((uint32_t) (-x) + (uint32_t) (-y)) <
				    (uint32_t) (-x))
#else
				   (((uint32_t) (-x) + (uint32_t) (-y))
				    > (uint32_t) INT32_MAX)
#endif
#endif
	     ))
	return UFLOW;

    return NFLOW;
}


static inline ALWAYS_INLINE int
catch_sub_bug(int32_t x, int32_t y)
{
    if (x > 0 && y < 0 && (
#if GREATER_NEGATIVE
			      (y == INT32_MIN) ||
#endif
#if GREATER_SIGNED
			      ((uint32_t) ((uint32_t) x + (uint32_t) (-y))
			       < (uint32_t) x)
#else
			      (((uint32_t) x + (uint32_t) (-y)) >
			       (uint32_t) INT32_MAX)
#endif
	))
	return OFLOW;

    else if (x < 0 && y > 0 && (
#if GREATER_NEGATIVE
				   (x == INT32_MIN) ||
#if GREATER_SIGNED
				   ((((uint32_t) (-x) + (uint32_t) y) != 0)
				    &&
				    ((uint32_t)
				     ((uint32_t) (-x) + (uint32_t) y) <
				     (uint32_t) (-x)))
#else
				   (((uint32_t) (-x) + (uint32_t) y) >
				    ((uint32_t) 1 + (uint32_t) INT32_MAX))
#endif
#else
#if GREATER_SIGNED
				   ((uint32_t)
				    ((uint32_t) (-x) + (uint32_t) y) <
				    (uint32_t) (-x))
#else
				   (((uint32_t) (-x) + (uint32_t) y) >
				    (uint32_t) INT32_MAX)
#endif
#endif
	     ))
	return UFLOW;

    return NFLOW;
}


/*
 * Instruction 0 Sum of two integers. It adds the value contained in the
 * `acc' with the value stored in a RAM word indexed by address `y'; the
 * result is stored in `acc'. The first operand is lost. 
 */
CPU_SYMBOL(vov_add)
{
    int             x;
    ram_t          *word;
    char            stra[FMT_ULONG];
    char            strb[FMT_ULONG];

    if (OPTION_TEST(VERBOSE)) {
	stra[stringify(stra, INSTR_ARGUMENT(vov.cpu.ir))] = '\0';
	strb[stringify(strb, vov.cpu.acc)] = '\0';

	msg("\t- adding opcode from offset ", stra,
	    " to accumulator: old value ", strb, " will be lost\n");
    }

    word = word_seek(INSTR_ARGUMENT(vov.cpu.ir), vov.ram);

    if (NULL == word) {
	stra[stringify(stra, INSTR_ARGUMENT(vov.cpu.ir))] = '\0';
	die(111,
	    "SIGSEGV received: Segmentation fault while reading offset ",
	    stra, "\n");
    }

    if (NFLOW != (x = catch_add_bug(vov.cpu.acc, word->opcode)))
	die(111, "SIGTERM received: ", ((x == OFLOW) ? "over" : "under"),
	    "flow detected\n");

    /*
     * update accumulator 
     */
    vov.cpu.acc += word->opcode;

    if (OPTION_TEST(VERBOSE)) {
	strb[stringify(strb, vov.cpu.acc)] = '\0';
	msg("\t- accumulator register now contains ", strb, "\n");
    }
}


/*
 * Instruction 1
 * Difference of two integers. It substracts the value kept in RAM[`y']
 * to the  value stored in  `acc'; the  result is stored  in `acc'. The
 * first operand is lost.
 */
CPU_SYMBOL(vov_sub)
{
    int             x;
    ram_t          *word;
    char            stra[FMT_ULONG];
    char            strb[FMT_ULONG];

    if (OPTION_TEST(VERBOSE)) {
	stra[stringify(stra, INSTR_ARGUMENT(vov.cpu.ir))] = '\0';
	strb[stringify(strb, vov.cpu.acc)] = '\0';

	msg("\t- subtracting opcode from offset ", stra,
	    " to accumulator: old value ", strb, " will be lost\n");
    }

    word = word_seek(INSTR_ARGUMENT(vov.cpu.ir), vov.ram);

    if (NULL == word) {
	stra[stringify(stra, INSTR_ARGUMENT(vov.cpu.ir))] = '\0';
	die(111,
	    "SIGSEGV received: Segmentation fault while reading offset ",
	    stra, "\n");
    }

    if (NFLOW != (x = catch_sub_bug(vov.cpu.acc, word->opcode)))
	die(111, "SIGTERM received: ", ((x == OFLOW) ? "over" : "under"),
	    "flow detected\n");

    /*
     * update accumulator 
     */
    vov.cpu.acc -= word->opcode;

    if (OPTION_TEST(VERBOSE)) {
	strb[stringify(strb, vov.cpu.acc)] = '\0';
	msg("\t- accumulator register now contains ", strb, "\n");
    }
}


/*
 * Instruction 2
 * Read of an  integer number from the input device. The  read value is
 * stored in the `acc'. The previously stored value is lost.
 */
CPU_SYMBOL(vov_read)
{
    char            stra[FMT_ULONG];
    long            x;

    if (OPTION_TEST(VERBOSE)) {
	stra[stringify(stra, vov.cpu.acc)] = '\0';
	msg("\t- reading new accumulator register value from input device: old value ", stra, " will be lost\n");
    }

    msg(">>> ");

    switch (get_stdin_line(&line)) {
    case -1:			/* error */
	die(111, "an error occurred while trying to read user input\n");

    case 0:			/* EOF */
	{
	    if (1 != exitasap)
		die(111, "got EOF.\n");

	    break;
	}

    default:
	{
	    longify(line.s, &x);
	    vov.cpu.acc = (int32_t) x;

	    if (OPTION_TEST(VERBOSE)) {
		stra[stringify(stra, vov.cpu.acc)] = '\0';
		msg("\t- accumulator register now contains ", stra, "\n");
	    }
	}
    }
}


/*
 * Instruction 3
 * Write of an integer value on  the output device. The value stored in
 * the `acc' is copied on the output device.  No value is lost.
 */
CPU_SYMBOL(vov_write)
{
    char            stracc[FMT_ULONG];

    if (OPTION_TEST(VERBOSE))
	msg("\t- printing the accumulator data register on the output device\n");

    stracc[stringify(stracc, vov.cpu.acc)] = '\0';
    msg(stracc, "\n");
}


/*
 * Instruction 4
 * Stores  the number  held  in the  `acc'  also in  RAM[`y']. Previous
 * content of RAM[`y'] is lost; content of `acc' isn't lost.
 */
CPU_SYMBOL(vov_store)
{
    char            stra[FMT_ULONG];

    if (OPTION_TEST(VERBOSE)) {
	stra[stringify(stra, INSTR_ARGUMENT(vov.cpu.ir))] = '\0';
	msg("\t- storing accumulator register on offset ", stra, "\n");
    }

    if (!add_word(INSTR_ARGUMENT(vov.cpu.ir), vov.cpu.acc, &vov.ram)) {
	stra[stringify(stra, INSTR_ARGUMENT(vov.cpu.ir))] = '\0';
	die(111,
	    "SIGSEGV received: Segmentation fault while writing on offset ",
	    stra, "\n");
    }
}


/*
 * Instruction 5
 * Copies the value held in  RAM[`y'] in the `acc'. Previous content of
 * `acc' is lost. Content of RAM[`y'] is kept.
 */
CPU_SYMBOL(vov_load)
{
    ram_t          *word;
    char            stra[FMT_ULONG];
    char            strb[FMT_ULONG];

    if (OPTION_TEST(VERBOSE)) {
	stra[stringify(stra, INSTR_ARGUMENT(vov.cpu.ir))] = '\0';
	strb[stringify(strb, vov.cpu.acc)] = '\0';

	msg("\t- loading opcode from offset ", stra,
	    " in the accumulator register: old value ", strb,
	    " will be lost\n");
    }

    word = word_seek(INSTR_ARGUMENT(vov.cpu.ir), vov.ram);

    if (NULL == word) {
	stra[stringify(stra, INSTR_ARGUMENT(vov.cpu.ir))] = '\0';
	die(111,
	    "SIGSEGV received: Segmentation fault while reading offset ",
	    stra, "\n");
    }

    vov.cpu.acc = word->opcode;

    if (OPTION_TEST(VERBOSE)) {
	strb[stringify(strb, vov.cpu.acc)] = '\0';
	msg("\t- accumulator register now contains ", strb, "\n");
    }
}


/*
 * Instruction 6
 * Usually called  a `program jump',  as it inserts an  interruption in
 * the  linear RAM  scan: it  makes the  next instruction  fetched from
 * RAM[`y']  instead of  the next  instruction  in RAM  after the  jump
 * instruction.
 */
CPU_SYMBOL(vov_jmp)
{
    char            stra[FMT_ULONG];
    char            strb[FMT_ULONG];

    if (OPTION_TEST(VERBOSE)) {
	stra[stringify(stra, INSTR_ARGUMENT(vov.cpu.ir))] = '\0';
	strb[stringify(strb, vov.cpu.pc)] = '\0';

	msg("\t- hop hop jump: set program counter register to ", stra,
	    ": old value ", strb, " will be lost\n");
    }

    vov.cpu.pc = INSTR_ARGUMENT(vov.cpu.ir);
}


/*
 * Instruction 7
 * Checks  the value  stored  in `acc'. If  this value  is  0, jump  as
 * `instruction 6',  otherwise don't  do anything  (a.k.a. `conditional
 * jump').
 */
CPU_SYMBOL(vov_jzero)
{
    char            stra[FMT_ULONG];
    char            strb[FMT_ULONG];

    if (OPTION_TEST(VERBOSE)) {
	stra[stringify(stra, vov.cpu.acc)] = '\0';
	msg("\t- conditional jump: testing accumulator register: ", stra,
	    " == 0\n");
    }

    if (0 == vov.cpu.acc) {
	if (OPTION_TEST(VERBOSE)) {
	    stra[stringify(stra, vov.cpu.pc)] = '\0';
	    strb[stringify(strb, INSTR_ARGUMENT(vov.cpu.ir))] = '\0';

	    msg("\t- test successfully: set program counter to ", strb,
		": old value ", stra, " will be lost\n");
	}

	vov.cpu.pc = INSTR_ARGUMENT(vov.cpu.ir);
    }

    else if (OPTION_TEST(VERBOSE)) {
	stra[stringify(stra, vov.cpu.pc)] = '\0';

	msg("\t- test failed: leave program counter register set to ",
	    stra, "\n");
    }
}


/*
 * Instruction 8
 * Halts the machine.
 */
CPU_SYMBOL(vov_halt)
{
    if (OPTION_TEST(VERBOSE))
	msg("\t- system halted\n");

    exitasap = 1;
}
