/***********************************************************************\
* vov is a tool that emulates the behavior of a Von-Neumann machine.    *
* Copyright (C) 2001-2008 Davide Scola <davide.scola@gmail.com>         *
*                                                                       *
* This file is part of vov.                                             *
*                                                                       *
* vov is free software; you can  redistribute it and/or modify it under *
* the terms of the GNU General  Public License as published by the Free *
* Software Foundation;  either version  2 of the  License, or  (at your *
* option) any later version.                                            *
*                                                                       *
* vov is  distributed in the hope  that it will be  useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or *
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License *
* for more details.                                                     *
*                                                                       *
* You should  have received a  copy of  the GNU General  Public License *
* along  with vov;  if  not,  write to  the  Free Software  Foundation, *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.    *
\***********************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#ifdef HAVE_CHANGE_LIMIT
# include <sys/time.h>
# include <sys/resource.h>
#endif

#include "line.h"
#include "buffer.h"
#include "human_out.h"
#include "utils.h"


/*****************************************************************\
*                           CREDITS                               *
*                                                                 *
* This func  is based on  libowfat (http://www.fefe.de/libowfat/) *
* written by Felix von Leitner and released under the GNU General *
* Public License published by the Free Software Foundation.       *
* In short, libowfat reimplements libdjb under GPL.               *
*                                                                 *
* Since I needed to change some little things and I don't want to *
* add any extra  no standard dependence to vov, I wrote this func *
\*****************************************************************/

/*
 * convert an unsigned long to a string 
 */
static inline ALWAYS_INLINE unsigned int
u_stringify(register char *str, register unsigned long num)
{
    register unsigned long len,
                    tmp,
                    len2;

    /*
     * first count the number of bytes needed 
     */
    for (len = 1, tmp = num; tmp > 9; ++len, tmp /= 10);

    if (NULL != str) {
	for (tmp = num, str += len, len2 = len + 1; --len2; tmp /= 10)
	    *--str = (tmp % 10) + '0';
    }

    return len;
}


/*****************************************************************\
*                           CREDITS                               *
*                                                                 *
* This func  is based on  libowfat (http://www.fefe.de/libowfat/) *
* written by Felix von Leitner and released under the GNU General *
* Public License published by the Free Software Foundation.       *
* In short, libowfat reimplements libdjb under GPL.               *
*                                                                 *
* Since I needed to change some little things and I don't want to *
* add any extra  no standard dependence to vov, I wrote this func *
\*****************************************************************/

/*
 * convert a signed long to a string 
 */
unsigned int
stringify(char *str, long int num)
{
    if (num < 0) {
	if (NULL != str)
	    *str++ = '-';

	return (u_stringify(str, -num) + 1);
    }

    return u_stringify(str, num);
}


/*****************************************************************\
*                           CREDITS                               *
*                                                                 *
* This func  is based on  libowfat (http://www.fefe.de/libowfat/) *
* written by Felix von Leitner and released under the GNU General *
* Public License published by the Free Software Foundation.       *
* In short, libowfat reimplements libdjb under GPL.               *
*                                                                 *
* Since I needed to change some little things and I don't want to *
* add any extra  no standard dependence to vov, I wrote this func *
\*****************************************************************/

/*
 * convert a string to an unsigned long int 
 */
void
ulongify(const char *src, unsigned long int *dest)
{
    register const char *tmp = src;
    register unsigned long int l = 0;
    register unsigned char c;
    register unsigned long int n;

    *dest = 0;			/* initialize dest */

    if (NULL != tmp) {
	while ((c = *tmp - '0') < 10) {
	    /*
	     * division is very slow on most architectures 
	     */
	    n = l << 3;

	    if ((n >> 3) != l)
		break;
	    if (n + (l << 1) < n)
		break;

	    n += l << 1;

	    if (n + c < n)
		break;

	    l = n + c;
	    ++tmp;
	}

	if (tmp - src)
	    *dest = l;
    }
}


/*****************************************************************\
*                           CREDITS                               *
*                                                                 *
* This func  is based on  libowfat (http://www.fefe.de/libowfat/) *
* written by Felix von Leitner and released under the GNU General *
* Public License published by the Free Software Foundation.       *
* In short, libowfat reimplements libdjb under GPL.               *
*                                                                 *
* Since I needed to change some little things and I don't want to *
* add any extra  no standard dependence to vov, I wrote this func *
\*****************************************************************/
void
longify(const char *src, long *dest)
{
    register const char *tmp = src;
    register long int l = 0;
    register unsigned char c;
    int             neg = 0;
    int             ok = 0;

    *dest = 0;			/* initialize dest */

    if (NULL != tmp) {
	switch (*tmp) {
	case '-':
	    neg = 1;
	case '+':
	    ++tmp;
	}

	while ((c = *tmp - '0') < 10) {
	    l = l * 10 + c;
	    ++tmp;
	    ok = 1;
	}

	if (ok)
	    *dest = (neg ? -l : l);
    }
}


/*****************************************************************\
*                           CREDITS                               *
*                                                                 *
* This func  is based on  libowfat (http://www.fefe.de/libowfat/) *
* written by Felix von Leitner and released under the GNU General *
* Public License published by the Free Software Foundation.       *
* In short, libowfat reimplements libdjb under GPL.               *
*                                                                 *
* Since I needed to change some little things and I don't want to *
* add any extra  no standard dependence to vov, I wrote this func *
\*****************************************************************/

/*
 * return the position of the first occurrence of needle
 * in haystack scanning only the first len characters of it.
 * If needle isn't found on the given chunk, the position
 * of the last checked character is returned.
 */
unsigned long
byte_chr(const void *haystack, unsigned long len, char needle)
{
    register char   c = needle;
    register const char *s = haystack;
    register const char *t = s + len;

    for (;;) {
	if (s == t)
	    break;
	if (*s == c)
	    break;
	++s;
	if (s == t)
	    break;
	if (*s == c)
	    break;
	++s;
	if (s == t)
	    break;
	if (*s == c)
	    break;
	++s;
	if (s == t)
	    break;
	if (*s == c)
	    break;
	++s;
    }

    return (s - (const char *) haystack);
}


/*****************************************************************\
*                           CREDITS                               *
*                                                                 *
* This func  is based on  libowfat (http://www.fefe.de/libowfat/) *
* written by Felix von Leitner and released under the GNU General *
* Public License published by the Free Software Foundation.       *
* In short, libowfat reimplements libdjb under GPL.               *
*                                                                 *
* Since I needed to change some little things and I don't want to *
* add any extra  no standard dependence to vov, I wrote this func *
\*****************************************************************/

/*
 * byte_copy copies in[0] to out[0], in[1] to out[1], ... and in[len-1] 
 * to out[len-1].
 */
void
byte_copy(void *out, unsigned long len, const void *in)
{
    register char  *s = out;
    register const char *t = in;
    register const char *u = t + len;

    if (len > 127) {
	while ((unsigned long) s & (sizeof(unsigned long) - 1)) {
	    if (t == u)
		break;
	    *s = *t;
	    ++s;
	    ++t;
	}

	/*
	 * s (destination) is now unsigned long aligned 
	 */
#ifndef __i386__
	if (!((unsigned long) t & (sizeof(unsigned long) - 1)))
#endif
	    while (t + sizeof(unsigned long) <= u) {
		*(unsigned long *) s = *(const unsigned long *) t;
		s += sizeof(unsigned long);
		t += sizeof(unsigned long);
	    }
    }

    for (;;) {
	if (t == u)
	    break;
	*s = *t;
	++s;
	++t;
	if (t == u)
	    break;
	*s = *t;
	++s;
	++t;
	if (t == u)
	    break;
	*s = *t;
	++s;
	++t;
	if (t == u)
	    break;
	*s = *t;
	++s;
	++t;
    }
}


/*****************************************************************\
*                           CREDITS                               *
*                                                                 *
* This func  is based on  libowfat (http://www.fefe.de/libowfat/) *
* written by Felix von Leitner and released under the GNU General *
* Public License published by the Free Software Foundation.       *
* In short, libowfat reimplements libdjb under GPL.               *
*                                                                 *
* Since I needed to change some little things and I don't want to *
* add any extra  no standard dependence to vov, I wrote this func *
\*****************************************************************/

/*
 * are you really looking for a comment here? :) 
 */
unsigned long
str_len(const char *in)
{
    register const char *t = in;

    for (;;) {
	if (!*t)
	    break;
	++t;
	if (!*t)
	    break;
	++t;
	if (!*t)
	    break;
	++t;
	if (!*t)
	    break;
	++t;
    }

    return (t - in);
}


#ifdef HAVE_CHANGE_LIMIT

/*
 * limit the given resource 
 */
static inline ALWAYS_INLINE void
set_limit(int x, unsigned long y)
{
    struct rlimit   z;
    char            stra[FMT_ULONG];

    if (-1 == getrlimit(x, &z)) {
	stra[stringify(stra, x)] = '\0';
	die(111, "impossible to get limit for resource ", stra, "\n");
    }

    z.rlim_cur = y;

    if (z.rlim_cur > z.rlim_max)
	z.rlim_cur = z.rlim_max;

    if (-1 == setrlimit(x, &z)) {
	stra[stringify(stra, x)] = '\0';
	die(111, "impossible to set limit for resource ", stra, "\n");
    }
}


/*
 * limit the memory that vov can eat 
 */
void
limit_memory(unsigned long max)
{
#ifdef RLIMIT_DATA
    set_limit(RLIMIT_DATA, max);
#endif

#ifdef RLIMIT_STACK
    set_limit(RLIMIT_STACK, max);
#endif

#ifdef RLIMIT_MEMLOCK
    set_limit(RLIMIT_MEMLOCK, max);
#endif

#ifdef RLIMIT_VMEM
    set_limit(RLIMIT_VMEM, max);
#endif

#ifdef RLIMIT_AS
    set_limit(RLIMIT_AS, max);
#endif
}

/*
 * limit the cpu time that can be used by vov; we don't want to run
 * forever 
 */
void
limit_cpu(unsigned long secs)
{
#ifdef RLIMIT_CPU
    set_limit(RLIMIT_CPU, secs);
#endif
}

#endif				/* ! HAVE_CHANGE_LIMIT */
